/*
** Copyright (C) 2002-2003 Ryan McCabe <ryan@numb.org>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License, version 2,
** as published by the Free Software Foundation.
*/

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <time.h>
#include <sys/socket.h>
#include <sys/un.h>

#include <snuff.h>
#include <snuff_util.h>

static void print_help_text(void);
static void print_version_info(void);
static int parse_timestr(char *str, u_int32_t *ttl);
static int connect_to_server(void);

static int get_alias_list(int sock);
static int delete_alias(int sock, char *user);
static int delete_all_aliases(int sock);
static int get_new_alias(int sock, time_t expire, char *user, size_t len);
static void snufferr(int code, char *buf, size_t len);

static int get_result(int sock);

int main(int argc, char **argv) {
	int opt;
	action_t action = ACTION_NONE;
	u_int32_t ttl;
	char *user = NULL;
	int sock;
	time_t now;

	while ((opt = getopt(argc, argv, "a:d:Dhlv")) != EOF) {
		switch (opt) {
			case 'a':
				action = ACTION_ADD;
				if (parse_timestr(optarg, &ttl) != 0) {
					debug("Invalid lifetime: %s", optarg);
					exit(-1);
				}
				time(&now);

				if ((time_t) ttl + now < now) {
					debug("The lifetime you specified is too large.");
					exit(-1);
				}

				break;

			case 'l':
				action = ACTION_LIST;
				break;

			case 'd':
				action = ACTION_DELETE;
				user = strdup(optarg);
				break;

			case 'D':
				action = ACTION_DELETE_ALL;
				break;

			case 'v':
				print_version_info();
				break;

			case 'h':
			default:
				print_help_text();
				break;
		}
	}

	if (action == ACTION_NONE)
		exit(0);

	sock = connect_to_server();
	if (sock == -1) {
		fprintf(stderr, "Unable to connect to the server\n");
		exit(-1);
	}

	if (action == ACTION_LIST) {
		if (get_alias_list(sock) != 0) {
			debug("Error listing current aliases.");
			exit(-1);
		}
	} else if (action == ACTION_DELETE) {
		if (delete_alias(sock, user) != 0) {
			debug("Error deleting the alias \"%s\"", user);
			free(user);
			exit(-1);
		}

		free(user);
	} else if (action == ACTION_DELETE_ALL) {
		if (delete_all_aliases(sock) != 0)
			exit(-1);
	} else if (action == ACTION_ADD) {
		char new_alias[MAX_USERLEN];
		int ret;

		ret = get_new_alias(sock, (time_t) ttl + now,
				new_alias, sizeof(new_alias));

		if (ret != 0) {
			debug("Error getting a new alias from the server");
			exit(-1);
		}

		printf("Your new email alias is %s\n", new_alias);
	}

	exit(0);
}

static void print_help_text(void) {
	printf("Usage: snuff [options]\n\
-a <time>   Request a new alias that will live for the specified dration.\n\
-d <alias>  Delete the specified alias.\n\
-D          Delete all aliases registered to you.\n\
-l          List all registered aliases for your email address\n\
-h          Print this help text.\n\
-v          Print version information.\n");
}

static void print_version_info(void) {
	printf("snuff version %s by %s\n", SNUFF_VERSION, SNUFF_AUTHOR);
}

static int parse_timestr(char *str, u_int32_t *ttl) {
	char *p;
	u_int32_t seconds = 0;

	while (1) {
		u_int32_t temp;

		if (*str == '\0')
			break;

		temp = strtoul(str, &p, 10);

		switch (tolower(*p)) {
			case 'y':
				temp *= 31536000;
				break;

			case 'w':
				temp *= 604800;
				break;

			case 'd':
				temp *= 86400;
				break;

			case 'h':
				temp *= 3600;
				break;

			case 'm':
				temp *= 60;
				break;

			case 's':
				break;

			case '\0':
				debug("Missing time unit");
				return (-1);

			default:
				debug("Invalid time unit: %c", *p);
				return (-1);
		}

		seconds += temp;
		str = p + 1;
	}

	*ttl = seconds;
	return (0);
}

static int connect_to_server(void) {
	int sock;
	struct sockaddr_un sun;

	sock = socket(PF_UNIX, SOCK_STREAM, 0);
	if (sock == -1) {
		debug("Error: socket: %s", strerror(errno));
		return (-1);
	}

	sun.sun_family = AF_UNIX;
	if (strlen(MSOCK_PATH) >= sizeof(sun.sun_path)) {
		debug("Path to sock (%s) is too long.", MSOCK_PATH);
		close(sock);
		return (-1);
	}

	strcpy(sun.sun_path, MSOCK_PATH);

	if (connect(sock, (struct sockaddr *) &sun, sizeof(sun)) != 0) {
		debug("Error: connect: %s", strerror(errno));
		close(sock);
		return (-1);
	}

	return (sock);
}

static int get_alias_list(int sock) {
	struct control_block control_block;
	int ret;
	int result;

	memset(&control_block, 0, sizeof(control_block));
	control_block.action = ACTION_LIST;

	ret = sock_write(sock, &control_block, sizeof(control_block));
	if (ret != sizeof(control_block)) {
		debug("Error: write: %s", strerror(errno));
		return (-1);
	}

	result = get_result(sock);
	if (result != RESULT_SUCCESS) {
		char buf[1024];

		snufferr(result, buf, sizeof(buf));
		fprintf(stderr, "Error listing aliases: %s\n", buf);
		return (-1);
	}

	while (1) {
		struct alias_block alias;

		ret = sock_read(sock, &alias, sizeof(alias));
		if (ret == -1)
			return (-1);

		if (ret == 0)
			break;

		if (ret != sizeof(alias)) {
			debug("Error: read: %s", strerror(errno));
			return (-1);
		}

		printf("user name: %s\t\texpire time: %s",
			alias.username, ctime(&alias.expire));
	}

	return (0);
}

static int delete_alias(int sock, char *user) {
	struct control_block control_block;
	int ret;
	int result;

	memset(&control_block, 0, sizeof(control_block));

	control_block.action = ACTION_DELETE;
	if (strlen(user) >= sizeof(control_block.username)) {
		debug("Username is too long.");
		return (-1);
	}

	strcpy(control_block.username, user);

	ret = sock_write(sock, &control_block, sizeof(control_block));
	if (ret != sizeof(control_block)) {
		debug("Error: write: %s", strerror(errno));
		return (-1);
	}

	result = get_result(sock);
	if (result != RESULT_SUCCESS) {
		char buf[1024];

		snufferr(result, buf, sizeof(buf));
		fprintf(stderr, "Error deleting alias: %s\n", buf);
		return (-1);
	}

	return (0);
}

static int delete_all_aliases(int sock) {
	struct control_block control_block;
	int ret;
	int result;

	memset(&control_block, 0, sizeof(control_block));

	control_block.action = ACTION_DELETE_ALL;

	ret = sock_write(sock, &control_block, sizeof(control_block));
	if (ret != sizeof(control_block)) {
		debug("Error: write: %s", strerror(errno));
		return (-1);
	}

	result = get_result(sock);
	if (result != RESULT_SUCCESS) {
		char buf[1024];

		snufferr(result, buf, sizeof(buf));
		fprintf(stderr, "Error deleting aliases: %s\n", buf);
		return (result);
	}

	return (0);
}

static int get_new_alias(int sock, time_t expire, char *user, size_t len) {
	struct control_block control_block;
	int ret;
	int result;

	memset(&control_block, 0, sizeof(control_block));

	control_block.action = ACTION_ADD;
	control_block.expire = expire;

	ret = sock_write(sock, &control_block, sizeof(control_block));
	if (ret != sizeof(control_block)) {
		debug("Error: write: %s", strerror(errno));
		return (-1);
	}

	result = get_result(sock);
	if (result != RESULT_SUCCESS) {
		char buf[1024];

		snufferr(result, buf, sizeof(buf));
		fprintf(stderr, "Error getting a new alias: %s\n", buf);
		return (-1);
	}

	ret = sock_read(sock, user, len - 1);
	if (ret < 1) {
		debug("Error reading username: %s", strerror(errno));
		return (-1);
	}
	user[ret] = '\0';

	return (0);
}

static int get_result(int sock) {
	int result;

	int ret = sock_read(sock, &result, sizeof(result));
	if (ret != sizeof(result)) {
		debug("Error: read: %s[%d]", strerror(errno), ret);
		return (RESULT_FAIL);
	}

	return (result);
}

static void snufferr(int code, char *buf, size_t len) {
	static const char *err_str[] = {
		"success",
		"unknown error",
		"operation failed",
		"argument specified is too long",
		"alias does not exist",
		"expiration is too soon",
		"you requested an invalid action",
		"you currently have no aliases",
		"you have reached the limit of allowed aliases"
	};

	if ((size_t) code <= sizeof(err_str) / sizeof(char *))
		xstrncpy(buf, err_str[code], len);
	else
		xstrncpy(buf, err_str[RESULT_UNKNOWN], len);
}
