/*
** Copyright (C) 2002-2003 Ryan McCabe <ryan@numb.org>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License, version 2,
** as published by the Free Software Foundation.
*/

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <pwd.h>
#include <grp.h>
#include <sys/types.h>
#include <sys/stat.h>

#include <snuff.h>
#include <snuff_util.h>

void *xmalloc(size_t len) {
	void *ret = malloc(len);
	if (ret == NULL) {
		debug("Out of memory\n");
		exit(-1);
	}

	return (ret);
}

void *xcalloc(size_t nmemb, size_t len) {
	void *ret = calloc(nmemb, len);

	if (ret == NULL) {
		debug("Out of memory\n");
		exit(-1);
	}

	return (ret);
}

char *xstrdup(const char *str) {
	char *ret = strdup(str);

	if (ret == NULL) {
		debug("Out of memory\n");
		exit(-1);
	}

	return (ret);
}

ssize_t sock_write(int sock, void *buf, size_t len) {
	ssize_t n;
	ssize_t written = 0;

	while (len > 0) {
		n = write(sock, buf, len);
		if (n == -1) {
			if (errno == EINTR)
				continue;
			return (-1);
		}

		written += n;
		len -= n;
		buf = (char *) buf + n;
	}

	return (written);
}

ssize_t sock_read(int sock, void *buf, size_t len) {
	int i;

	for (i = 0 ; i < 5 ; i++) {
		ssize_t ret = read(sock, buf, len);
		if (ret == -1) {
			if (errno == EINTR)
				continue;
			return (-1);
		}

		return (ret);
	}

	return (-1);
}

/*
** Copy at most n-1 characters from src to dest and nul-terminate dest.
** Returns a pointer to the destination string.
*/

char *xstrncpy(char *dest, const char *src, size_t n) {
	char *ret = dest;

	if (n == 0)
		return (dest);

	while (--n > 0 && (*dest++ = *src++) != '\0')
		;
	*dest = '\0';

	return (ret);
}

char *get_username(uid_t uid) {
	struct passwd *pw;
	char *ret;

	pw = getpwuid(uid);
	if (pw == NULL)
		return (NULL);

	ret = xstrdup(pw->pw_name);

	return (ret);
}

void xsleep(size_t seconds) {
	struct timeval tv;

	tv.tv_usec = 0;
	tv.tv_sec = seconds;

	select(0, NULL, NULL, NULL, &tv);
}

/*
** Go background.
*/

int go_background(void) {
	int fd;
	int ret;

	ret = fork();
	if (ret == -1)
		return (-1);

	if (ret != 0)
		_exit(0);

	if (setsid() == -1) {
		debug("setsid: %s", strerror(errno));
		return (-1);
	}

	if (chdir("/") != 0) {
		debug("chdir: %s", strerror(errno));
		return (-1);
	}

	umask(022);

	fd = open("/dev/null", O_RDWR);
	if (fd == -1) {
		debug("open: /dev/null: %s", strerror(errno));
		return (-1);
	}

	dup2(fd, 0);
	dup2(fd, 1);
	dup2(fd, 2);

	return (0);
}

/*
** Find the user specified by "temp_user"
** Returns 0 on success, -1 on failure.
*/

int find_user(const char *temp_user, uid_t *uid) {
	struct passwd *pw;

	pw = getpwnam(temp_user);
	if (pw == NULL) {
		char *end;
		u_int32_t temp_uid = strtoul(temp_user, &end, 10);

		if (*end != '\0')
			return (-1);
		else
			*uid = temp_uid;
	} else
		*uid = pw->pw_uid;

	return (0);
}

/*
** Find the group specified by "temp_group"
** Returns 0 on success, -1 on failure.
*/

int find_group(const char *temp_group, gid_t *gid) {
	struct group *gr;

	gr = getgrnam(temp_group);
	if (gr == NULL) {
		char *end;
		u_int32_t temp_gid = strtoul(temp_group, &end, 10);

		if (*end != '\0')
			return (-1);
		else
			*gid = temp_gid;
	} else
		*gid = gr->gr_gid;

	return (0);
}

/*
** Drop privileges and run with specified UID/GID.
** Returns 0 on success, -1 on failure.
*/

int drop_privs(uid_t new_uid, gid_t new_gid) {

	if (new_gid != getgid()) {
		if (setgid(new_gid) != 0) {
			debug("setgid(%u): %s", new_gid, strerror(errno));
			return (-1);
		}
	}

	if (new_uid != getuid()) {
		struct passwd *pw;
		gid_t my_gid;
		int ret;

		pw = getpwuid(new_uid);
		if (pw == NULL) {
			debug("getpwuid(%u): No such user", new_uid);
			return (-1);
		}

		if (new_gid != getgid())
			my_gid = new_gid;
		else
			my_gid = pw->pw_gid;

		ret = initgroups(pw->pw_name, my_gid);

		if (ret != 0) {
			debug("initgroups(%s, %u): %s", pw->pw_name, my_gid,
				strerror(errno));
			return (-1);
		}

		if (setuid(new_uid) != 0) {
			debug("setuid(%u): %s", new_uid, strerror(errno));
			return (-1);
		}
	}

	return (0);
}
